
/**
 * Module dependencies.
 */

var fs = require('fs')
  , path = require('path')
  , debug = require('debug')('package-tree')
  , dirname = path.dirname
  , join = path.join
  , resolve = path.resolve
  , exists = fs.existsSync
  , paths = process.env.NODE_PATH;

/**
 * Cache.
 */

var cache = {};

/**
 * Expose `tree()`.
 */

exports = module.exports = tree;

// split paths

if (paths) paths = paths.split(':');

/**
 * Return the next "node_modules" dir in `dir`.
 *
 * @return {String}
 * @api private
 */

exports.nextModuleDir = function(dir){
  return join(dir.split('node_modules').slice(0, -2).join('node_modules'), 'node_modules');
};

/**
 * Return the directories to attempt,
 * including those in NODE_PATH.
 *
 * @param {Type} name
 * @return {Type}
 * @api public
 */

exports.dirs = function(dir){
  var dirs = [].concat(paths).concat(dir);
  while (!~dirs.indexOf(dir = exports.nextModuleDir(dir))) {
    dirs.push(dir);
  }
  return dirs;
};

/**
 * Return package tree for the given package.json `file`.
 *
 * Example:
 *
 *    var tree = require('package-tree');
 *    console.log(tree('/mymodule/package.json'));
 *
 * @param {String} file
 * @return {Object}
 * @api public
 */

function tree(file){
  if (cache[file]) return cache[file];
  var pkg = JSON.parse(fs.readFileSync(file, 'utf8'));
  var deps = pkg.dependencies;
  var bundled = pkg.bundledDependencies || pkg.bundleDependencies;
  pkg.dirname = dirname(file);
  pkg.deps = [];

  if (!deps && !bundled) return pkg;
  var depNames = Object.keys(deps || {}).concat(bundled || []);

  depNames.forEach(function(dep){
    var dirs = exports.dirs(join(pkg.dirname, 'node_modules'));
    var ret;

    for (var i = 0, len = dirs.length; i < len; ++i) {
      ret = load(dep, dirs[i]);
      if (ret) break;
    }

    if (!ret) throw new Error('Failed to resolve package "' + pkg.name + '"\'s dep "' + dep + '"');

    pkg.deps.push(ret);
  });

  function load(dep, dir) {
    var file = join(dir, dep, 'package.json');

    debug('trying %s', file);
    if (exists(file)) {
      debug('found %s', file);
      return tree(file);
    }
  }

  cache[file] = pkg;
  return pkg;
};